/* todo (a date-based reminder)
 * Copyright (c) 2002, 2004, Emil Mikulic.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * -=-
 *
 * format: mm dd Text text text...
 *
 * e.g.:   12 25 Christmas Day
 *         # comments work too
 *
 * $Id: todo.c,v 1.6 2004/03/08 06:51:15 emikulic Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define DEFAULT_DAYS 14
#define POOLSIZE 100
#define TODOFILE /* $HOME/ */ ".todo"
static char default_editor[] = "vi";



typedef struct {
	time_t when;
	char *what;
} todo_entry;

static todo_entry *todo_list = NULL;
static int todo_pool = 0, todo_used = 0;
static time_t now, today;
static char *todo_fn;



static void *
xmalloc(const size_t s) {
	void *tmp = malloc(s);

	if (!tmp) {
		fprintf(stderr, "Out of memory!\n");
		exit(EXIT_FAILURE);
	}
	return tmp;
}



static int
getyear(void)
{
	struct tm *t;

	t = localtime(&now);
	return t->tm_year; /* + 1900 */
}



static void
growpool(void)
{
	void *tmp;

	todo_pool += POOLSIZE;
	tmp = realloc(todo_list, todo_pool * sizeof(todo_entry));

	if (!tmp) {
		fprintf(stderr, "Out of memory in growpool()\n");
		exit(EXIT_FAILURE);
	}

	todo_list = (todo_entry *)tmp;
}



#define LINEBUF 256
static char line[LINEBUF];
static int linenum = 0;

static void
getline(FILE *f)
{
	int i;
	char c;

	linenum++;
	for (i=0; i<LINEBUF-1 && (c=(char)fgetc(f)) != '\n' && !feof(f); i++)
		line[i] = c;
	line[i] = 0;

	if (i == LINEBUF-1) {
		fprintf(stderr, "Line too long at %s:%d\n", todo_fn, linenum);
		exit(EXIT_FAILURE);
	}
}



static void
addentry(const int month, const int day, const char *what)
{
	struct tm t;
	todo_entry *e;
	
	if (todo_used == todo_pool) growpool();
	e = &(todo_list[todo_used]);
	todo_used++;

	t.tm_mday = day;
	t.tm_mon = month - 1;
	t.tm_year = getyear();
	t.tm_sec = 0;
	t.tm_min = 0;
	t.tm_hour = 0;
	t.tm_isdst = 0;

	e->when = mktime(&t);
	e->what = strdup(what);
	if (e->when < today) {
		t.tm_year++;
		e->when = mktime(&t);
	}
}



static int
parseline(void)
{
	int left, pos, month = 0, day = 0;

	if (!strlen(line)) return 1;
	if (line[0] == '#') return 1;

	for (pos = 0; line[pos] == ' ' && line[pos]; pos++); /* whitespace */
	if (!line[pos]) return 0;
	if (line[pos] == '#') return 1;

	for (left = pos; line[pos] != ' ' && line[pos]; pos++); /* numbers */
	if (!line[pos]) return 0;

	line[pos] = 0;
	month = atoi(line+left);

	for (pos++; line[pos] == ' ' && line[pos]; pos++); /* whitespace */
	if (!line[pos]) return 0;

	for (left = pos; line[pos] != ' ' && line[pos]; pos++); /* numbers */
	if (!line[pos]) return 0;

	line[pos] = 0;
	day = atoi(line+left);

	for (pos++; line[pos] == ' ' && line[pos]; pos++); /* whitespace */

	if (
		month < 1 || month > 12 ||
		day < 1 || day > 31 ||
		!strlen(line+pos)
	)
		return 0;

	addentry(month, day, line+pos);
	return 1;
}



static int
todocmp(const void *t1, const void *t2)
{
	const todo_entry *e1 = (const todo_entry *)t1;
	const todo_entry *e2 = (const todo_entry *)t2;

	if (e1->when < e2->when) return -1; else
	if (e1->when > e2->when) return  1; else
	return 0;
}



static void
loadtodo(void)
{
	FILE *f = fopen(todo_fn, "r");

	if (!f) {
		fprintf(stderr, "Can't open %s\n", todo_fn);
		exit(EXIT_FAILURE);
	}

	while (!feof(f)) {
		getline(f);
		if (!parseline())
			fprintf(stderr, "Malformed line in %s:%d.\n",
				todo_fn, linenum);
	}
	fclose(f);

	/* Ooh. */
	qsort(todo_list, todo_used, sizeof(todo_entry), todocmp);
}



static void
do_usage(const char *me)
{
	printf(
		"usage: %s [-a|-e|<n>]\n"
		"\n"
		"\t-a\tdisplay all reminders\n"
		"\t-e\tedit reminder file (%s)\n"
		"\t<n>\tshow the next <n> days\n"
		"\n",
		me, todo_fn);
	exit(EXIT_SUCCESS);
}



static void
showdays(const int limit_days)
{
	int i = 0;

	while (i<todo_used) {
		struct tm *t;
		todo_entry *e = &(todo_list[i]);
		int days = (e->when - today) / 86400;

		if (days > limit_days && limit_days != -1) return;

		t = localtime(&(e->when));
		printf("(%02d/%02d) [+%3d days] %s\n",
			t->tm_mon+1, t->tm_mday,
			days,
			e->what);

		i++;
	}
}



static void
do_edit(void)
{
	char *cmd, *editor = getenv("EDITOR");

	if (!editor || !strlen(editor)) editor = default_editor;

	cmd = (char *)xmalloc(strlen(editor) + strlen(todo_fn) + 2);
	sprintf(cmd, "%s %s", editor, todo_fn);
	system(cmd);
	free(cmd);
	exit(EXIT_SUCCESS);
}



int
main(int argc, char **argv)
{
	int days = DEFAULT_DAYS;
	char *home = getenv("HOME");
	struct tm t1, *t2;

	now = time(NULL);

	t2 = localtime(&now);
	t1.tm_isdst = 0;
	t1.tm_sec = 0;
	t1.tm_min = 0;
	t1.tm_hour = 0;
	t1.tm_mday = t2->tm_mday;
	t1.tm_mon = t2->tm_mon;
	t1.tm_year = t2->tm_year;
	today = mktime(&t1);

	todo_fn = (char *)xmalloc(strlen(home) + strlen(TODOFILE) + 2);
	sprintf(todo_fn, "%s/%s", home, TODOFILE);

	if (argc > 2) do_usage(argv[0]);
	if (argc == 2) {
		if (strcmp(argv[1], "-e") == 0) do_edit(); else
		if (strcmp(argv[1], "-a") == 0) days = -1; else
		{
			char *tmp = NULL;
			days = (int)strtol(argv[1], &tmp, 10);

			if (tmp != argv[1] + strlen(argv[1]))
				do_usage(argv[0]);
		}
	}

	loadtodo();
	showdays(days);
	return 0;
}
